import numpy as np
import pylab as plt
from qunfold.utils import compute_chi2


################### PLOTTING CONFIGURATIONS ###################
# https://matplotlib.org/stable/gallery/color/named_colors.html
label2color = {
    "Truth": "tab:blue",
    "Measured": "tab:orange",
    "MI": "purple",  # matrix inversion
    "IBU": "red",  # iterative bayesian
    "SA": "green",  # simulated annealing
    "HYB": "dodgerblue",  # hybrid solver
    "QA": "orchid",  # quantum annealing
}
default_color = "black"
figsize = (12, 9)
chi2_ndigits = 2
alpha = 0.25
marker = "o"
markersize = 2
linewidth = 1
elinewidth = 1
capsize = 3
capthick = 1
ticks_fontsize = 8
labels_fontsize = 14
legend_fontsize = 10
###############################################################


class QPlotter:
    def __init__(self, response, measured, truth, unfolded, covariance, binning, method, ybottom=0.0, norm=False):
        self.response = response[1:-1, 1:-1]
        self.measured = measured[1:-1]
        self.truth = truth[1:-1]
        self.unfolded = unfolded[1:-1]
        self.covariance = covariance[1:-1, 1:-1]
        self.binning = binning[1:-1]
        self.method = method
        self.ybottom = ybottom
        self.norm = norm

    @staticmethod
    def histogram_plot(ax, xedges, hist, label, ylabel="Entries", ybottom=0.0, norm=False):
        if norm:
            hist = hist / np.sum(hist)
            ylabel = "Frequency"
        hist = np.append(hist, [hist[-1]])
        color = label2color.get(label, default_color)
        ax.step(xedges, hist, label=label, color=color, where="post", linewidth=linewidth)
        ax.fill_between(xedges, hist, color=color, alpha=alpha, step="post")
        ax.set_ylabel(ylabel, fontsize=labels_fontsize)
        ax.set_ylim(ybottom)

    @staticmethod
    def errorbar_plot(ax, xmid, hist, err, xlims, label, chi2, norm=False):
        if norm:
            err = err / np.sum(hist)
            hist = hist / np.sum(hist)
        color = label2color.get(label, default_color)
        rchi2 = round(chi2, ndigits=chi2_ndigits)
        label = rf"Unfolded {label} ($\chi^2 = {rchi2}$)"
        ax.errorbar(
            x=xmid,
            y=hist,
            yerr=err,
            label=label,
            color=color,
            marker=marker,
            ms=markersize,
            elinewidth=elinewidth,
            capsize=capsize,
            capthick=capthick,
            linestyle="None",
        )
        ax.tick_params(labelsize=ticks_fontsize, top=False, right=False, labelbottom=False, reset=True)
        ax.legend(fontsize=legend_fontsize)
        ax.set_xlim(left=xlims[0], right=xlims[1])
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)

    @staticmethod
    def ratio_plot(ax, xmid, ratio, err, label, xticks, xlabel="Bins"):
        ax.axhline(y=1, color=label2color["Truth"], linewidth=linewidth)
        color = label2color.get(label, default_color)
        ax.errorbar(
            x=xmid,
            y=ratio,
            yerr=err,
            color=color,
            marker=marker,
            ms=markersize,
            elinewidth=elinewidth,
            capsize=capsize,
            capthick=capthick,
            linestyle="None",
        )
        ax.tick_params(labelsize=ticks_fontsize, right=False, reset=True)
        ax.set_xticks(xticks)
        ax.set_xlabel(xlabel, fontsize=labels_fontsize)
        ax.set_ylabel("Ratio to\nTruth", fontsize=labels_fontsize)
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)

    def _plot_response(self):
        fig, ax = plt.subplots(figsize=figsize)
        xs = ys = self.binning
        R = self.response
        heatmap = ax.pcolormesh(xs, ys, R.T)
        ax.set_xticks(xs)
        ax.set_yticks(ys)
        ax.tick_params(labelsize=ticks_fontsize, reset=True)
        ax.set_xlabel("Measured", fontsize=labels_fontsize)
        ax.set_ylabel("Truth", fontsize=labels_fontsize)
        fig.colorbar(heatmap, ax=ax)
        fig.tight_layout()

    def _plot_histograms(self):
        fig = plt.figure(figsize=figsize)
        gs = fig.add_gridspec(nrows=2, ncols=1, height_ratios=[3, 1], hspace=0)
        ax1 = fig.add_subplot(gs[0])
        ax2 = fig.add_subplot(gs[1], sharex=ax1)

        measured, truth = self.measured, self.truth
        sol, cov = self.unfolded, self.covariance
        err = np.sqrt(np.diag(cov))
        sol_ratio = sol / truth
        err_ratio = err / truth
        label = self.method
        widths = np.diff(self.binning)
        xmid = self.binning[:-1] + 0.5 * widths
        xlims = (self.binning[0], self.binning[-1])
        chi2 = compute_chi2(observed=sol, expected=truth)

        self.histogram_plot(
            ax=ax1, xedges=self.binning, hist=truth, label="Truth", ybottom=self.ybottom, norm=self.norm
        )
        self.histogram_plot(
            ax=ax1, xedges=self.binning, hist=measured, label="Measured", ybottom=self.ybottom, norm=self.norm
        )
        self.errorbar_plot(ax=ax1, xmid=xmid, hist=sol, err=err, xlims=xlims, label=label, chi2=chi2, norm=self.norm)
        yticks = [tick for tick in ax1.get_yticks() if tick != 0]
        ax1.set_yticks(yticks)
        self.ratio_plot(ax=ax2, xmid=xmid, ratio=sol_ratio, err=err_ratio, label=label, xticks=self.binning)
        fig.tight_layout()

    def show_response(self):
        self._plot_response()
        plt.show()
        plt.close()

    def save_response(self, path):
        self._plot_response()
        plt.savefig(path)
        plt.close()

    def show_histograms(self):
        self._plot_histograms()
        plt.show()
        plt.close()

    def save_histograms(self, path):
        self._plot_histograms()
        plt.savefig(path)
        plt.close()



import os
import sys
import numpy as np
import scipy as sp
import dimod
import minorminer
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor
from dwave.samplers import SimulatedAnnealingSampler
from dwave.samplers import SteepestDescentSolver
from dwave.system import LeapHybridSampler
from dwave.system import DWaveSampler, FixedEmbeddingComposite

try:
    import gurobipy
except ImportError:
    pass


class QUnfolder:
    def __init__(self, response, measured, binning, lam=0.0):
        self.R = response
        self.d = measured
        self.binning = binning
        self.lam = lam
        self.sol_pick = "weighted-average"

    @property
    def num_bins(self):
        return len(self.d)

    @property
    def num_logical_qubits(self):
        return self.dwave_bqm.num_variables

    @property
    def num_physical_qubits(self):
        return sum(len(chain) for chain in self.graph_embedding.values())

    @property
    def num_bits(self):
        eff = np.sum(self.R, axis=0)
        eff[np.isclose(eff, 0)] = 1
        exp = np.ceil(self.d / eff)
        return [(int(np.ceil(np.log2(x * 1.2))) if x else 1) for x in exp]

    @property
    def precision_vectors(self):
        num_bits = self.num_bits
        pvecs = [2 ** np.arange(num_bits[i]) for i in range(self.num_bins)]
        return pvecs

    @property
    def linear_coeffs(self):
        return -2 * (self.R.T @ self.d)

    @property
    def quadratic_coeffs(self):
        L = self._get_laplacian()
        return (self.R.T @ self.R) + self.lam * (L.T @ L)

    def _get_laplacian(self):
        x = (self.binning[:-1] + self.binning[1:]) / 2
        n = len(x)
        L = np.zeros(shape=(n, n))
        for i in range(1, n - 1):
            h1 = x[i] - x[i - 1]
            h2 = x[i + 1] - x[i]
            L[i, i - 1] = 2 / (h1 * (h1 + h2))
            L[i, i] = -2 / (h1 * h2)
            L[i, i + 1] = 2 / (h2 * (h1 + h2))
        L[0, :] = L[1, :]
        L[-1, :] = L[-2, :]
        L = 2 * L / np.max(np.abs(L))
        return L

    def _get_qubo_matrix(self):
        dim = self.num_bins
        a = self.linear_coeffs
        B = self.quadratic_coeffs
        pvecs = self.precision_vectors
        linear_blocks = [a[i] * np.diag(pvecs[i]) for i in range(dim)]
        quadratic_blocks = [[B[i, j] * np.outer(pvecs[i], pvecs[j]) for j in range(dim)] for i in range(dim)]
        return sp.linalg.block_diag(*linear_blocks) + np.block(quadratic_blocks)

    def _get_dwave_bqm(self):
        Q = self.qubo_matrix
        size = len(Q)
        linear = {i: Q[i, i] for i in range(size)}
        quadratic = {(i, j): 2 * Q[i, j] for i in range(size) for j in range(i + 1, size)}
        return dimod.BinaryQuadraticModel(linear, quadratic, vartype=dimod.BINARY)

    def _decode_array(self, arr):
        split_indices = np.cumsum(self.num_bits[:-1])
        bitstrings_list = np.split(arr, indices_or_sections=split_indices)
        decoded_arr = np.array([pvec @ bits for pvec, bits in zip(self.precision_vectors, bitstrings_list)])
        return decoded_arr

    def _decode_matrix(self, mat):
        split_indices = np.cumsum(self.num_bits[:-1])
        rows = cols = np.split(np.arange(len(mat)), indices_or_sections=split_indices)
        pvecs = self.precision_vectors
        nbins = self.num_bins
        decoded_mat = np.zeros(shape=(nbins, nbins))
        for i in range(nbins):
            for j in range(nbins):
                block = mat[np.ix_(rows[i], cols[j])]
                decoded_mat[i, j] = pvecs[i] @ block @ pvecs[j]
        return decoded_mat

    def _post_process_sampleset(self, sampleset):
        sampleset = SteepestDescentSolver().sample(self.dwave_bqm, initial_states=sampleset)
        solutions = np.array([rec.sample for rec in sampleset.record])
        energies = np.array([rec.energy for rec in sampleset.record])
        if self.sol_pick == "mean":
            binsol = np.mean(solutions, axis=0)
            devs = solutions - binsol
            bincov = (devs.T @ devs) / len(devs)
        elif self.sol_pick == "weighted-average":
            temperature = (np.max(energies) - np.min(energies)) / np.log(len(energies))
            weights = np.exp(-(1 / temperature) * (energies - np.min(energies)))
            weights /= np.sum(weights)
            binsol = np.average(solutions, weights=weights, axis=0)
            devs = solutions - binsol
            bincov = (devs.T @ (devs * weights[:, np.newaxis])) / (1 - np.sum(weights**2))
        elif self.sol_pick == "lowest-energy":
            raise NotImplementedError  # TODO
            from qunfold.utils import approx_hessian

            binsol = solutions[np.argmin(energies)]
            hess = approx_hessian(f=self.dwave_bqm.energy, x=binsol)
            bincov = np.linalg.pinv(hess)
        sol = self._decode_array(arr=binsol)
        cov = self._decode_matrix(mat=bincov)
        return sol, cov

    def _get_graph_embedding(self, **kwargs):
        source_edgelist = list(self.dwave_bqm.quadratic) + list((v, v) for v in self.dwave_bqm.linear)
        target_edgelist = self._sampler.edgelist
        return minorminer.find_embedding(S=source_edgelist, T=target_edgelist, **kwargs)

    def _run_montecarlo_toys(self, num_toys, prog_bar, num_cores, **kwargs):
        def run_toy(_):
            smeared_d = np.random.poisson(self.d)
            toy = QUnfolder(self.R, smeared_d, binning=self.binning, lam=self.lam)
            toy.initialize_qubo_model()
            if isinstance(self._sampler, DWaveSampler):
                embedding = toy._get_graph_embedding()
                sampler = FixedEmbeddingComposite(self._sampler, embedding=embedding)
            else:
                sampler = self._sampler
            sampleset = sampler.sample(toy.dwave_bqm, **kwargs)
            sol, _ = toy._post_process_sampleset(sampleset)
            return sol

        max_workers = num_cores if num_cores is not None else os.cpu_count()
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            jobs = executor.map(run_toy, range(num_toys))
            desc = "Running MC toys"
            disable = not prog_bar
            results = list(tqdm(jobs, total=num_toys, desc=desc, disable=disable))
        cov_toys = np.cov(results, rowvar=False)
        return cov_toys

    def initialize_qubo_model(self):
        self.qubo_matrix = self._get_qubo_matrix()
        self.dwave_bqm = self._get_dwave_bqm()

    def solve_simulated_annealing(self, num_reads, num_toys=None, num_cores=None, seed=None):
        self._sampler = SimulatedAnnealingSampler()
        sampleset = self._sampler.sample(self.dwave_bqm, num_reads=num_reads, seed=seed)
        sol, cov = self._post_process_sampleset(sampleset)
        if num_toys is not None:
            cov_toys = self._run_montecarlo_toys(num_toys, num_cores, num_reads=num_reads, seed=seed)
            cov += cov_toys
        return sol, cov

    def solve_hybrid_sampler(self):
        self._sampler = LeapHybridSampler()
        sampleset = self._sampler.sample(self.dwave_bqm)
        sample = sampleset.record[0].sample
        sol = self._decode_array(arr=sample)
        cov = np.diag(sol)
        return sol, cov

    def set_quantum_device(self, device_name=None, dwave_token=None):
        self._sampler = DWaveSampler(solver=device_name, token=dwave_token)

    def set_graph_embedding(self, **kwargs):
        self.graph_embedding = self._get_graph_embedding(**kwargs)

    def solve_quantum_annealing(self, num_reads, num_toys=None, prog_bar=True, num_cores=None):
        sampler = FixedEmbeddingComposite(self._sampler, embedding=self.graph_embedding)
        sampleset = sampler.sample(self.dwave_bqm, num_reads=num_reads)
        sol, cov = self._post_process_sampleset(sampleset)
        if num_toys is not None:
            cov_toys = self._run_montecarlo_toys(num_toys, prog_bar, num_cores, num_reads=num_reads)
            cov += cov_toys
        return sol, cov

    def compute_energy(self, x):
        xbin = []
        num_bits = self.num_bits
        for i in range(self.num_bins):
            bitstr = bin(int(x[i]))[2:].zfill(num_bits[i])
            xbin.extend(map(int, bitstr[::-1]))
        energy = self.dwave_bqm.energy(sample=xbin)
        return energy

    if "gurobipy" in sys.modules:

        def solve_gurobi_integer(self):
            model = gurobipy.Model()
            model.setParam("OutputFlag", 0)
            x = [model.addVar(vtype=gurobipy.GRB.INTEGER, lb=0, ub=2**b - 1) for b in self.num_bits]
            a = self.linear_coeffs
            B = self.quadratic_coeffs
            model.setObjective(a @ x + x @ B @ x, sense=gurobipy.GRB.MINIMIZE)
            model.optimize()
            sol = np.array([var.x for var in x])
            cov = np.diag(sol)
            return sol, cov

        def solve_gurobi_binary(self):
            model = gurobipy.Model()
            model.setParam("OutputFlag", 0)
            num_bits = self.num_bits
            x = [model.addVar(vtype=gurobipy.GRB.BINARY) for i in range(self.num_bins) for _ in range(num_bits[i])]
            Q = self.qubo_matrix
            model.setObjective(x @ Q @ x, sense=gurobipy.GRB.MINIMIZE)
            model.optimize()
            bitstr = np.array([var.x for var in x], dtype=int)
            arrays = np.split(bitstr, np.cumsum(num_bits[:-1]))
            sol = np.array([int("".join(arr.astype(str))[::-1], base=2) for arr in arrays], dtype=float)
            cov = np.diag(sol)
            return sol, cov



import sympy
import numpy as np
from hypothesis import settings, given
from hypothesis import strategies as st
from qunfold.utils import approx_hessian


deg = st.integers(min_value=0, max_value=4)
xi = st.integers(min_value=0, max_value=1e6)


@settings(deadline=None)
@given(degrees=st.tuples(deg, deg, deg), point=st.tuples(xi, xi, xi))
def test_approx_hessian(degrees, point):
    varlist = [sympy.Symbol(f"x{i}") for i in range(len(degrees))]
    poly_sympy = 0.0
    for deg, var in zip(degrees, varlist):
        for exp in range(deg + 1):
            coeff = np.random.rand() * 5 - 2
            poly_sympy += coeff * var**exp
    n = len(varlist)
    hess_sympy = sympy.Matrix(np.zeros(shape=(n, n)))
    for i in range(n):
        for j in range(n):
            hess_sympy[i, j] = sympy.diff(poly_sympy, varlist[i], varlist[j])
    func = sympy.lambdify(args=varlist, expr=poly_sympy, modules="numpy")
    hess = sympy.lambdify(args=varlist, expr=hess_sympy, modules="numpy")
    hessian1 = approx_hessian(f=lambda x: func(*x), x=point)
    hessian2 = hess(*point)
    assert np.allclose(hessian1, hessian2)



import sys
import jax
import numpy as np
import scipy as sp
from tqdm import tqdm
from qunfold import QUnfolder

try:
    import gurobipy
except ImportError:
    pass


def normalize_response(response, truth_mc):
    iszero = truth_mc == 0.0
    response /= np.where(iszero, 1.0, truth_mc)
    diag = np.diag(response)
    response[np.diag_indices_from(response)] = np.where(iszero, 1.0, diag)
    return response


def compute_chi2(observed, expected):
    iszero = expected == 0
    observed = np.where(iszero, observed + 1, observed)
    expected = np.where(iszero, expected + 1, expected)
    chi2 = np.sum((observed - expected) ** 2 / expected)
    chi2_red = chi2 / (len(expected) - 1)
    return chi2_red


def approx_hessian(f, x):
    x = jax.numpy.array(x, dtype=float)
    n = len(x)
    hessian = jax.numpy.zeros(shape=(n, n))
    grad_f = jax.grad(f)
    for i in range(n):
        v = jax.numpy.zeros_like(x)
        v = v.at[i].set(1.0)
        _, hvp = jax.jvp(grad_f, primals=(x,), tangents=(v,))
        hessian = hessian.at[:, i].set(hvp)
    return hessian


def lambda_optimizer(response, measured, truth, binning, num_reps=30, verbose=False, seed=None):
    if "gurobipy" not in sys.modules:
        raise ModuleNotFoundError("Function 'lambda_optimizer' requires Gurobi solver")
    np.random.seed(seed)

    def objective_fun(lam):
        unfolder = QUnfolder(response, measured, binning=binning, lam=lam)
        unfolder.initialize_qubo_model()
        sol, _ = unfolder.solve_gurobi_integer()
        obs = sol[1:-1]
        exp = truth[1:-1]
        chi2 = compute_chi2(observed=obs, expected=exp)
        return chi2

    best_lam = 0
    min_fun = objective_fun(best_lam)
    options = {"xatol": 0, "maxiter": 100, "disp": 3 if verbose else 0}
    for _ in tqdm(range(num_reps), desc="Optimizing lambda"):
        bounds = (0, np.random.rand())
        minimizer = sp.optimize.minimize_scalar(fun=objective_fun, method="bounded", bounds=bounds, options=options)
        lam = minimizer.x
        fun = minimizer.fun
        if fun < min_fun:
            best_lam = lam
            min_fun = fun
    return best_lam
